
#include <stdio.h>
#include <stdlib.h>
#include "mt.h"

extern void	Error();

#define RATIO	(27.0/256.0)	/* ==pow(3,3)/pow(4,4) */

static long	*Link;

static long	Trace(old)
long		old;
{
	long	new=Link[old];

	return (new<0)?old:(Link[old]=Trace(new));
}

static int	X,Y,Z;
static long	N,**Last;

#define SITES	((double)X*(double)Y*(double)Z)

static void	RandomSitePercolation(p)
double		p;
{
	char	*tlc="too large cluster.";
	int	y,x,z,iy,ix;
	long	l,m,links=1,clusters=0;
	double	sites,total1=0.0,total2=0.0,largest=0.0;

#define CHECK(size)	if ((size)>=0) Error(tlc)

	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) Last[y][x]=0;
	for (z=0; z<Z; z++)
	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) {
	    if (MT_dRandom()>=p)
		l=0;
	    else if ((l=Last[y][x])!=0) {
		CHECK(--Link[l=Trace(l)]);
/*
		if ((m=Last[y-1][x])!=0) {
		    if ((m=Trace(m))!=l) {
			CHECK(Link[l]+=Link[m]); Link[m]=l;
		    }
		    if ((m=Last[y][x-1])!=0 && (m=Trace(m))!=l) {
			CHECK(Link[l]+=Link[m]); Link[m]=l;
		    }
		}
		else if ((m=Last[y][x-1])!=0 && m!=l) {
		    CHECK(Link[l]+=Link[m]); Link[m]=l;
		}
*/
		if ((m=Last[y-1][x])!=0 && (m=Trace(m))!=l) {
		    CHECK(Link[l]+=Link[m]); Link[m]=l;
		}
		if ((m=Last[y][x-1])!=0 && (m=Trace(m))!=l) {
		    CHECK(Link[l]+=Link[m]); Link[m]=l;
		}
/**/
	    }
	    else if ((l=Last[y-1][x])!=0) {
		CHECK(--Link[l=Trace(l)]);
		if ((m=Last[y][x-1])!=0 && m!=l) {
		    CHECK(Link[l]+=Link[m]); Link[m]=l;
		}
	    }
	    else if ((l=Last[y][x-1])!=0) {
		CHECK(--Link[l]);
	    }
	    else {
		if (links==N) {
		    for (iy=0; iy<Y; iy++)
		    for (ix=0; ix<X; ix++)
			if ((l=Last[iy][ix])!=0) Last[iy][ix]=Trace(l);

		    for (links=l=1; l<N; l++)
			if (Link[l]<0) ++links; else Link[l]=0;

		    if (links==N) Error("too many links.");
#ifdef	DEBUG
		    (void)fprintf(stderr,"%ld\n",links);
#endif
		    for (iy=0; iy<Y; iy++)
		    for (ix=0; ix<X; ix++)
			if ((l=Last[iy][ix])!=0) {
			    Last[iy][ix]=Link[l];
			    Link[l]=(long)iy*(long)X+(long)ix+1;
			}

		    for (links=l=1; l<N; l++)
			if ((m=Link[l])!=0) {
			    while (m>0) {
				--m; m=Last[iy=(int)(m/X)][ix=(int)(m%X)];
				Last[iy][ix]=links;
			    }
			    Link[links++]=m;
			}
		}
		Link[l=links++]=(-1);
	    }
	    Last[y][x]=l;
	}
	for (l=1; l<links; l++)
	    if ((sites=(-(double)Link[l]))>0.0) {
		++clusters; total1+=sites; total2+=(sites*sites);
		if (sites>largest) largest=sites;
	    }

	(void)printf("%lf\t%ld\t%ld\t",total1/SITES,links,clusters);
	if (clusters==0)
	    (void)puts("0\t0");
	else
	    (void)printf("%lf\t%lf\n",
			 largest/total1,(total2-largest*largest)/total1);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	double		n,p;
	unsigned	seed;
	int		x,y;

	if (argc<5)
	    Error("usage : cubic sites_x sites_y sites_z seed {probabilities}");

	if ((X=atoi(*(++argv)))<=0 ||
	    (Y=atoi(*(++argv)))<=0 ||
	    (Z=atoi(*(++argv)))<=0 ||
	    (n=RATIO*SITES)>=(double)((size_t)(~0)/sizeof(long)) ||
	    (N=(long)(n+1.0))<=0) Error("bad number of sites.");

	if (sscanf(*(++argv),"%u",&seed)!=1 || seed==0)
	    Error("bad random number seed.");

#define ALLOC(type,size)	(type *)malloc(sizeof(type)*(size))

	if ((Link=ALLOC(long,(size_t)N))==NULL ||
	    (Last=ALLOC(long *,(size_t)(1+Y)))==NULL ||
	    (Last[0]=ALLOC(long,(size_t)X+(size_t)Y*(size_t)(1+X)))==NULL)
	    Error("memory allocation error.");

	++Last;
	for (x=0; x<X; x++) Last[-1][x]=0;
	for (y=0; y<Y; y++) {
	    Last[y]=Last[y-1]+1+X; Last[y][-1]=0;
	}
	MT_sRandom(seed);

	if (argc==5)
	    while (scanf("%lf",&p)==1) RandomSitePercolation(p);
	else
	    while (--argc>=5) RandomSitePercolation(atof(*(++argv)));

	return 0;
}
