
#include <stdio.h>
#include <stdlib.h>
#include "mt.h"

extern void	Error();

#define RATIO	(4.0/27.0)	/* ==pow(2,2)/pow(3,3) */

static long	*Link;

static long	Trace(old)
long		old;
{
	long	new=Link[old];

	return (new<0)?old:(Link[old]=Trace(new));
}

static int	X,Y;
static long	N,*Last;

#define SITES	((double)X*(double)Y)

static void	RandomSitePercolation(p)
double		p;
{
	char	*tlc="too large cluster.";
	int	x,y,ix;
	long	l,m,links=1,clusters=0;
	double	sites,total1=0.0,total2=0.0,largest=0.0;

#define CHECK(size)	if ((size)>=0) Error(tlc)

	for (x=0; x<X; x++) Last[x]=0;
	for (y=0; y<Y; y++)
	for (x=0; x<X; x++) {
	    if (MT_dRandom()>=p)
		l=0;
	    else if ((l=Last[x])!=0) {
		CHECK(--Link[l=Trace(l)]);
		if ((m=Last[x-1])!=0 && m!=l) {
		    CHECK(Link[l]+=Link[m]); Link[m]=l;
		}
	    }
	    else if ((l=Last[x-1])!=0) {
		CHECK(--Link[l]);
	    }
	    else {
		if (links==N) {
		    for (ix=0; ix<X; ix++)
			if ((l=Last[ix])!=0) Last[ix]=Trace(l);

		    for (links=l=1; l<N; l++)
			if (Link[l]<0) ++links; else Link[l]=0;

		    if (links==N) Error("too many links.");
#ifdef	DEBUG
		    (void)fprintf(stderr,"%ld\n",links);
#endif
		    for (ix=0; ix<X; ix++)
			if ((l=Last[ix])!=0) {
			    Last[ix]=Link[l]; Link[l]=(long)ix+1;
			}

		    for (links=l=1; l<N; l++)
			if ((m=Link[l])!=0) {
			    while (m>0) {
				m=Last[ix=(int)(m-1)]; Last[ix]=links;
			    }
			    Link[links++]=m;
			}
		}
		Link[l=links++]=(-1);
	    }
	    Last[x]=l;
	}
	for (l=1; l<links; l++)
	    if ((sites=(-(double)Link[l]))>0.0) {
		++clusters; total1+=sites; total2+=(sites*sites);
		if (sites>largest) largest=sites;
	    }

	(void)printf("%lf\t%ld\t%ld\t",total1/SITES,links,clusters);
	if (clusters==0)
	    (void)puts("0\t0");
	else
	    (void)printf("%lf\t%lf\n",
			 largest/total1,(total2-largest*largest)/total1);
}

int	main(argc,argv)
int	argc;
char	**argv;
{
	double		n,p;
	unsigned	seed;

	if (argc<4)
	    Error("usage : square sites_x sites_y seed {probabilities}");

	if ((X=atoi(*(++argv)))<=0 ||
	    (Y=atoi(*(++argv)))<=0 ||
	    (n=RATIO*SITES)>=(double)((size_t)(~0)/sizeof(long)) ||
	    (N=(long)(n+1.0))<=0) Error("bad number of sites.");

	if (sscanf(*(++argv),"%u",&seed)!=1 || seed==0)
	    Error("bad random number seed.");

	if ((Link=(long *)malloc(sizeof(long)*(size_t)N))==NULL ||
	    (Last=(long *)malloc(sizeof(long)*(1+(size_t)X)))==NULL)
	    Error("memory allocation error.");

	++Last; Last[-1]=0;

	MT_sRandom(seed);

	if (argc==4)
	    while (scanf("%lf",&p)==1) RandomSitePercolation(p);
	else
	    while (--argc>=4) RandomSitePercolation(atof(*(++argv)));

	return 0;
}
